/* GLSL_DoF2: Depth of Field effect with GLSL shaders and FBOs
   Copyright (C) 2008 Angelo "Encelo" Theodorou
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <stdio.h>
#include <stdlib.h>

#include <SDL/SDL.h>
#include <GL/gl.h>
#include <GL/glext.h> 

#include "init.h"
#include "textfile.h"


void InitFBOs(struct FBO *fbo, int width, int height)
{
	int i;
	GLenum status;

	glGenFramebuffersEXT(1, &(fbo->frame[0]));
	glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo->frame[0]);

	glGenRenderbuffersEXT(1, &(fbo->depth));
	glBindRenderbufferEXT(GL_RENDERBUFFER_EXT, fbo->depth);
	glRenderbufferStorageEXT(GL_RENDERBUFFER_EXT, GL_DEPTH_COMPONENT24, width, height);
	glFramebufferRenderbufferEXT(GL_FRAMEBUFFER_EXT, GL_DEPTH_ATTACHMENT_EXT, GL_RENDERBUFFER_EXT, fbo->depth);

	glGenTextures(1, &fbo->texid[0]);

	glBindTexture(GL_TEXTURE_2D, fbo->texid[0]);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F_ARB,  width, height, 0, GL_RGBA, GL_FLOAT, NULL);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, fbo->texid[0], 0);

	status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
	if (status != GL_FRAMEBUFFER_COMPLETE_EXT) {
		printf("FBO[0] Error: 0x%x\n", status);
		exit(-1);
	}


	for (i = 1; i < NUM_FBOS; i++) {
		glGenFramebuffersEXT(1, &(fbo->frame[i]));
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo->frame[i]);

		glGenTextures(1, &fbo->texid[i]);

		glBindTexture(GL_TEXTURE_2D, fbo->texid[i]);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA16F_ARB, width/2, height/2, 0, GL_RGBA, GL_FLOAT, NULL);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glFramebufferTexture2DEXT(GL_FRAMEBUFFER_EXT, GL_COLOR_ATTACHMENT0_EXT, GL_TEXTURE_2D, fbo->texid[i], 0);

		status = glCheckFramebufferStatusEXT(GL_FRAMEBUFFER_EXT);
		if (status != GL_FRAMEBUFFER_COMPLETE_EXT) {
			printf("FBO[%d] Error: 0x%x\n", i, status);
			exit(-1);
		}
	}
}


void DeleteFBOs(struct FBO *fbo)
{
	glDeleteTextures(NUM_FBOS, fbo->texid);;
	glDeleteRenderbuffersEXT(1, &(fbo->depth));
	glDeleteFramebuffersEXT(NUM_FBOS, fbo->frame);
}



void InitShaders(struct Shaders *sh)
{
	int i;
	char *vs, *fs, *filename;
	const char *vv, *ff;
	const char *vfiles[NUM_SHADERS] = {"first.vert", "second.vert", "third.vert", "fourth.vert", "fifth.vert", "fifth.vert"};
	const char *ffiles[NUM_SHADERS] = {"first.frag", "second.frag", "third.frag", "fourth.frag", "fifth.frag", "renderblur.frag"};

	for (i = 0; i < NUM_SHADERS; i++) {

		sh->p[i] = glCreateProgram();

		filename = (char *)malloc(sizeof(char) * (strlen(SHADERS_DIR) + strlen(vfiles[i])));
		strcpy(filename, SHADERS_DIR);
		strcat(filename, vfiles[i]);

		sh->vert[i] = glCreateShader(GL_VERTEX_SHADER);
		vs = textFileRead(filename);
		printf("Vert[%d]: %s\n", i, filename);
		free(filename);
		vv = vs;
		glShaderSource(sh->vert[i], 1, &vv, NULL);
		free(vs);
		glCompileShader(sh->vert[i]);
		glAttachShader(sh->p[i], sh->vert[i]);

		filename = (char *)malloc(sizeof(char) * (strlen(SHADERS_DIR) + strlen(ffiles[i])));
		strcpy(filename, SHADERS_DIR);
		strcat(filename, ffiles[i]);

		sh->frag[i] = glCreateShader(GL_FRAGMENT_SHADER);
		fs = textFileRead(filename);
		printf("Frag[%d]: %s\n", i, filename);
		free(filename);
		ff = fs;
		glShaderSource(sh->frag[i], 1, &ff, NULL);
		free(fs);
		glCompileShader(sh->frag[i]);
		glAttachShader(sh->p[i], sh->frag[i]);

		glLinkProgram(sh->p[i]);
		printObjectInfoLog(sh->p[i]);
	}

	printf("\n");
}


void DeleteShaders(struct Shaders *sh)
{
	int i;

	for (i = 0; i < NUM_SHADERS; i++) {
		glDetachShader(sh->p[i], sh->frag[i]);
		glDeleteShader(sh->frag[i]);
		glDetachShader(sh->p[i], sh->vert[i]);
		glDeleteShader(sh->vert[i]);
		glDeleteProgram(sh->p[i]);
	}
}


GLuint LoadTexture(char *file)
{
	GLuint id;
	SDL_Surface *image;

	if (!(image = (SDL_Surface *)IMG_Load(file))) {
		printf("Texture %s load error", file);
		exit(-1);
	}

	glGenTextures(1, &id);
	glBindTexture(GL_TEXTURE_2D, id);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	if ((image->format)->Amask)
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, image->w, image->h, 0, GL_RGBA, GL_UNSIGNED_BYTE, image->pixels);
	else
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB8, image->w, image->h, 0, GL_RGB, GL_UNSIGNED_BYTE, image->pixels);
	SDL_FreeSurface(image);

	return id;
}
