/* GLSL_DoF2: Depth of Field effect with GLSL shaders and FBOs
   Copyright (C) 2008 Angelo "Encelo" Theodorou
 
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
  
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include <SDL/SDL.h>
#include <GL/gl.h>
#include <GL/glext.h> 

#include "init.h"
#include "extra.h"
#include "textfile.h"

#define DEFAULT_DISTANCE 9.0f
#define DEFAULT_RANGE 6.0f

void DrawQuad(void);


int main(int argc, char **argv)
{
	SDL_Surface *screen;
	SDL_Event event;
	const SDL_VideoInfo* info = NULL;
	int width = 640;
	int height = 480;
	int bpp = 0;
	int flags = 0;

	const GLubyte* string;
	GLint value;

	GLuint texid[2];
	SDL_Surface *image;

	struct FBO fbo;
	struct Shaders sh;
	GLint loc;

	int i;
	int quit = 0;
	Uint32 interval;
	GLfloat movz = 0.0f, roty = 0.0f;
	int move = 0, rotate = 1, bg = 1, rendermode = 0;
	float focalDistance = DEFAULT_DISTANCE;
	float focalRange = DEFAULT_RANGE;

/* ----- SDL init --------------- */
	if (SDL_Init(SDL_INIT_VIDEO) < 0) {
		fprintf(stderr, "Video initialization failed: %s\n", SDL_GetError());
		exit(-1);
	}

	atexit(SDL_Quit);
	
	info = SDL_GetVideoInfo();
	bpp = info->vfmt->BitsPerPixel;

	
/* ----- OpenGL attribute setting via SDL --------------- */
	SDL_GL_SetAttribute(SDL_GL_RED_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_GREEN_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_BLUE_SIZE, 5);
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER, 1);
	
/*	flags = SDL_OPENGL | SDL_FULLSCREEN; */
	flags = SDL_OPENGL;

	
/* ----- Setting up the screen surface --------------- */
	if ((screen = SDL_SetVideoMode(width, height, bpp, flags)) == 0) {
		fprintf(stderr, "Video mode set failed: %s\n", SDL_GetError());
        	exit(-1);
	}

	SDL_WM_SetCaption("GLSL_DoF2 by Encelo", NULL);


/* ----- Checking for OpenGL 2 --------------- */
	string = glGetString(GL_VENDOR);
	printf("Vendor: %s\n", string);
	string = glGetString(GL_RENDERER);
	printf("Renderer: %s\n", string);
	string = glGetString(GL_VERSION);
	printf("OpenGL Version: %s\n", string);
	if (string[0] != '2') {
		printf("OpenGL 2 not supported!\n");
		exit(-1);
	}
	string = glGetString(GL_SHADING_LANGUAGE_VERSION);
	printf("GLSL Version: %s\n\n", string);


/* ----- OpenGL init --------------- */
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glEnable(GL_DEPTH_TEST);

	InitFBOs(&fbo, width, height);
	InitShaders(&sh);

	glUseProgram(sh.p[2]);
	loc = glGetUniformLocation(sh.p[2], "Width");
	glUniform1i(loc, width*2);
	glUseProgram(sh.p[3]);
	loc = glGetUniformLocation(sh.p[3], "Height");
	glUniform1i(loc, height*2);
	glUseProgram(sh.p[4]);
	loc = glGetUniformLocation(sh.p[4], "Tex0");
	glUniform1i(loc, 0);
	loc = glGetUniformLocation(sh.p[4], "Tex1");
	glUniform1i(loc, 1);

	texid[0] = LoadTexture("tex0.jpg");
	texid[1] = LoadTexture("tex1.png");


/* ----- Event cycle --------------- */
	while (!quit) {
		while (SDL_PollEvent(&event)) {
			switch (event.type) {
			case SDL_QUIT:
				quit = 1;
				break;
			case SDL_KEYDOWN:
				switch( event.key.keysym.sym ){
				case SDLK_ESCAPE:
				case SDLK_q:
					quit = 1;
					break;
				case SDLK_F1:
					SDL_WM_ToggleFullScreen(screen);
					break;
				case SDLK_SPACE:
					move = !move;
					break;
				case SDLK_r:
					rotate = !rotate;
					break;
				case SDLK_b:
					bg = !bg;
					break;
				case SDLK_m:
					rendermode = !rendermode;
					break;
				case SDLK_p:
					printf("focalDistance = %f\n", focalDistance);
					printf("focalRange = %f\n", focalRange);
					break;
				case SDLK_KP7:
					focalDistance -= 0.5f;
					break;
				case SDLK_KP8:
					focalDistance = DEFAULT_DISTANCE;
					break;
				case SDLK_KP9:
					focalDistance += 0.5f;
					break;
				case SDLK_KP4:
					focalRange -= 0.5f;
					break;
				case SDLK_KP5:
					focalRange = DEFAULT_RANGE;
					break;
				case SDLK_KP6:
					focalRange += 0.5f;
					break;
				case SDLK_KP0:
					focalDistance = DEFAULT_DISTANCE;
					focalRange = DEFAULT_RANGE;
					movz = 0.0f;
					roty = 0.0f;
					break;
				case SDLK_1:
					focalDistance = 9.0f;
					break;
				case SDLK_2:
					focalDistance = 12.5f;
					break;
				case SDLK_3:
					focalDistance = 17.5f;
					break;
				case SDLK_4:
					focalDistance = 21.0f;
					break;
				case SDLK_5:
					focalDistance = 25.0f;
					break;
				}
			}
		}

		interval = FrameTiming();

		
/* ----- Blitting on the screen --------------- */
		if (move)
			movz += 0.0025 * interval;
		if (rotate)
			roty += 0.1 * interval;

		/*
		 * fbo.frame[0] -> Scene rendering
		 * fbo.frame[1] -> Original downsampled scene rendering
		 * fbo.frame[2] -> X Filtered downsampled scene rendering
		 * fbo.frame[3] -> XY Filtered downsampled scene rendering
		*/

		/* First pass: scene rendering */
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo.frame[0]);
		glDrawBuffer(GL_COLOR_ATTACHMENT0_EXT);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glActiveTexture(GL_TEXTURE0);
		glUseProgram(sh.p[0]);
		loc = glGetUniformLocation(sh.p[0], "focalDistance");
		glUniform1f(loc, focalDistance);
		loc = glGetUniformLocation(sh.p[0], "focalRange");
		glUniform1f(loc, focalRange);

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		gluPerspective(45.0f, (GLfloat)width/(GLfloat)height, 0.1f, 100.0f);
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();
		glTranslatef(0.0f, -2.0f, -10.0f);
		glRotatef(15.0f, 1.0f, 0.0f, 0.0f);

		/* Brick room */
		glBindTexture(GL_TEXTURE_2D, texid[0]);
		if (bg) {
			glPushMatrix();
			glTranslatef(0.0f, 6.0f, -10.0f);
			glScalef(10.0f, 10.0f, 10.0f);
			DrawCube(1);
			glPopMatrix();
		}

		/* Rotating cubes */
		glBindTexture(GL_TEXTURE_2D, texid[1]);
		for (i = 0; i < 5; i++) {
			glPushMatrix();
			glTranslatef(0.0f, 0.0f, 3.0 * sinf(movz));
			glTranslatef(-2.0f, 0.0f, 0.0);
			glTranslatef(i * 2.0f, 0.0f, i * (-4.0));
			glRotatef(roty, 0.0f, 1.0f, 0.0f);
			DrawCube(0);
			glPopMatrix();
		}

		glMatrixMode(GL_PROJECTION);
		glLoadIdentity();
		glMatrixMode(GL_MODELVIEW);
		glLoadIdentity();


		if (!rendermode) {
		/* Second pass: downsampling */
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo.frame[1]);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glBindTexture(GL_TEXTURE_2D, fbo.texid[0]);
		glUseProgram(sh.p[1]);

		glViewport(0, 0, width/2, height/2);

		DrawQuad();


		/* Third pass: Gaussian filtering along the X axis */
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo.frame[2]);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glBindTexture(GL_TEXTURE_2D, fbo.texid[1]);
		glUseProgram(sh.p[2]);

		DrawQuad();


		/* Fourth pass: Gaussian filtering along the Y axis */
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, fbo.frame[3]);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glBindTexture(GL_TEXTURE_2D, fbo.texid[2]);
		glUseProgram(sh.p[3]);

		DrawQuad();


		/* Fifth pass: final compositing */
		glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
		glDrawBuffer(GL_BACK);
		glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, fbo.texid[0]);
		glActiveTexture(GL_TEXTURE1);
		glBindTexture(GL_TEXTURE_2D, fbo.texid[3]);
		glUseProgram(sh.p[4]);

		glViewport(0, 0, width, height);

		glBegin(GL_QUADS);
			glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 0.0f);
			glMultiTexCoord2f(GL_TEXTURE1, 0.0f, 0.0f);
			glVertex2d(-1, -1);
			glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 0.0f);
			glMultiTexCoord2f(GL_TEXTURE1, 1.0f, 0.0f);
			glVertex2d(1, -1);
			glMultiTexCoord2f(GL_TEXTURE0, 1.0f, 1.0f);
			glMultiTexCoord2f(GL_TEXTURE1, 1.0f, 1.0f);
			glVertex2d(1, 1);
			glMultiTexCoord2f(GL_TEXTURE0, 0.0f, 1.0f);
			glMultiTexCoord2f(GL_TEXTURE1, 0.0f, 1.0f);
			glVertex2d(-1, 1);
		glEnd();
		}

		else { /* Rendering the blur channel only */
			glBindFramebufferEXT(GL_FRAMEBUFFER_EXT, 0);
			glDrawBuffer(GL_BACK);
			glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
			glBindTexture(GL_TEXTURE_2D, fbo.texid[0]);
			glUseProgram(sh.p[5]);

			DrawQuad();
		}

		SDL_GL_SwapBuffers();		
		/* SDL_Delay(20); */ /* Decomment this if you want 1/50th screen update */
	}

	
/* ----- Quitting --------------- */
	glDeleteTextures(2, texid);
	DeleteShaders(&sh);
	DeleteFBOs(&fbo);

	SDL_Quit();
	return 0;	
}


void DrawQuad(void)
{
	glBegin(GL_QUADS);
		glTexCoord2f(0.0f, 0.0f);
		glVertex2d(-1, -1);
		glTexCoord2f(1.0f, 0.0f);
		glVertex2d(1, -1);
		glTexCoord2f( 1.0f, 1.0f);
		glVertex2d(1, 1);
		glTexCoord2f(0.0f, 1.0f);
		glVertex2d(-1, 1);
	glEnd();
}
